/*
 * (C) Copyright 2013 Samsung Electronics Co. Ltd
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <common.h>
#include <asm/arch/cpu.h>
#include "pmic.h"
#include "setup.h"

#define rGPG0CON	*(volatile unsigned long *)(0x14000080)
#define rGPG0DAT	*(volatile unsigned long *)(0x14000084)
#define rGPG0PUD	*(volatile unsigned long *)(0x14000088)

#define rGPC1CON	*(volatile unsigned long *)(0x13410020)
#define rGPC1DAT	*(volatile unsigned long *)(0x13410024)
#define rGPC1PUD	*(volatile unsigned long *)(0x13410028)
#define rGPC1DRV	*(volatile unsigned long *)(0x1341002c)

board_id_t get_board_info(void) {
	board_id_t id = { .is_valid = 0 };

	id.is_valid = 1;

	/* some RC Settle time.. */
	sdelay(100000);

	/* read values */
	uint32_t value = rGPG0DAT;

	/* disable pull-ups/downs to save static power */
	rGPG0PUD = 0;

	/* parse readings */
	/*
	 * Despite what the Dev and Proto Schematics say, the EVT schematic
	 * is correct:
	 * A floating BOARD_ID pin (e.g., value now reads 0) is a form-factor
	 * A high BOARD_ID in is a Dev board.
	 */
	id.is_dev_board = !!(value & (1 << 1));

	/*
	 * Board-Rev:
	 * 0x7 is Proto
	 * 0x6 is EVT
	 * 0x5 is DVT
	 * 0x4 is Cert/DVT2
	 * 0b0-- is Production.
	 */
	id.revision = ((value & 0x1C) >> 2) & 0x7;

	return id;
}

void Delay(void)
{
	unsigned long i;
	for (i = 0; i < DELAY; i++)
		;
}

void IIC0_SCLH_SDAH(void)
{
	IIC0_ESCL_Hi;
	IIC0_ESDA_Hi;
	Delay();
}

void IIC0_SCLH_SDAL(void)
{
	IIC0_ESCL_Hi;
	IIC0_ESDA_Lo;
	Delay();
}

void IIC0_SCLL_SDAH(void)
{
	IIC0_ESCL_Lo;
	IIC0_ESDA_Hi;
	Delay();
}

void IIC0_SCLL_SDAL(void)
{
	IIC0_ESCL_Lo;
	IIC0_ESDA_Lo;
	Delay();
}

void IIC0_ELow(void)
{
	IIC0_SCLL_SDAL();
	IIC0_SCLH_SDAL();
	IIC0_SCLH_SDAL();
	IIC0_SCLL_SDAL();
}

void IIC0_EHigh(void)
{
	IIC0_SCLL_SDAH();
	IIC0_SCLH_SDAH();
	IIC0_SCLH_SDAH();
	IIC0_SCLL_SDAH();
}

void IIC0_EStart(void)
{
	IIC0_SCLH_SDAH();
	IIC0_SCLH_SDAL();
	Delay();
	IIC0_SCLL_SDAL();
}

void IIC0_EEnd(void)
{
	IIC0_SCLL_SDAL();
	IIC0_SCLH_SDAL();
	Delay();
	IIC0_SCLH_SDAH();
}

void IIC0_EAck_write(void)
{
	unsigned long ack;

	/* Function <- Input */
	IIC0_ESDA_INP;

	IIC0_ESCL_Lo;
	Delay();
	IIC0_ESCL_Hi;
	Delay();
	ack = GPIO0_DAT;
	IIC0_ESCL_Hi;
	Delay();
	IIC0_ESCL_Hi;
	Delay();

	/* Function <- Output (SDA) */
	IIC0_ESDA_OUTP;

	ack = (ack >> GPIO0_DAT_SHIFT) & 0x1;

	IIC0_SCLL_SDAL();
}

void IIC0_EAck_read(void)
{
	/* Function <- Output */
	IIC0_ESDA_OUTP;

	IIC0_ESCL_Lo;
	IIC0_ESCL_Lo;
	IIC0_ESDA_Hi;
	IIC0_ESCL_Hi;
	IIC0_ESCL_Hi;
	/* Function <- Input (SDA) */
	IIC0_ESDA_INP;

	IIC0_SCLL_SDAL();
}

void IIC0_ESetport(void)
{
	/* Pull Up/Down Disable	SCL, SDA */
	DIS_GPIO0_PUD;

	IIC0_ESCL_Hi;
	IIC0_ESDA_Hi;

	/* Function <- Output (SCL) */
	IIC0_ESCL_OUTP;
	/* Function <- Output (SDA) */
	IIC0_ESDA_OUTP;

	Delay();
}

void IIC0_EWrite(unsigned char ChipId,
		unsigned char IicAddr, unsigned char IicData)
{
	unsigned long i;

	IIC0_EStart();

	/* write chip id */
	for (i = 7; i > 0; i--) {
		if ((ChipId >> i) & 0x0001)
			IIC0_EHigh();
		else
			IIC0_ELow();
	}

	/* write */
	IIC0_ELow();

	/* ACK */
	IIC0_EAck_write();

	/* write reg. addr. */
	for (i = 8; i > 0; i--) {
		if ((IicAddr >> (i-1)) & 0x0001)
			IIC0_EHigh();
		else
			IIC0_ELow();
	}

	/* ACK */
	IIC0_EAck_write();

	/* write reg. data. */
	for (i = 8; i > 0; i--) {
		if ((IicData >> (i-1)) & 0x0001)
			IIC0_EHigh();
		else
			IIC0_ELow();
	}

	/* ACK */
	IIC0_EAck_write();

	IIC0_EEnd();
}

void IIC0_ERead(unsigned char ChipId,
		unsigned char IicAddr, unsigned char *IicData)
{
	unsigned long i, reg;
	unsigned char data = 0;

	IIC0_EStart();

	/* write chip id */
	for (i = 7; i > 0; i--) {
		if ((ChipId >> i) & 0x0001)
			IIC0_EHigh();
		else
			IIC0_ELow();
	}

	/* write */
	IIC0_ELow();

	/* ACK */
	IIC0_EAck_write();

	/* write reg. addr. */
	for (i = 8; i > 0; i--) {
		if ((IicAddr >> (i-1)) & 0x0001)
			IIC0_EHigh();
		else
			IIC0_ELow();
	}

	/* ACK */
	IIC0_EAck_write();

	IIC0_EStart();

	/* write chip id */
	for (i = 7; i > 0; i--) {
		if ((ChipId >> i) & 0x0001)
			IIC0_EHigh();
		else
			IIC0_ELow();
	}

	/* read */
	IIC0_EHigh();
	/* ACK */
	IIC0_EAck_write();

	/* read reg. data. */
	IIC0_ESDA_INP;

	IIC0_ESCL_Lo;
	IIC0_ESCL_Lo;
	Delay();

	for (i = 8; i > 0; i--) {
		IIC0_ESCL_Lo;
		IIC0_ESCL_Lo;
		Delay();
		IIC0_ESCL_Hi;
		IIC0_ESCL_Hi;
		Delay();
		reg = GPIO0_DAT;
		IIC0_ESCL_Hi;
		IIC0_ESCL_Hi;
		Delay();
		IIC0_ESCL_Lo;
		IIC0_ESCL_Lo;
		Delay();

		reg = (reg >> GPIO0_DAT_SHIFT) & 0x1;

		data |= reg << (i-1);
	}

	/* ACK */
	IIC0_EAck_read();
	IIC0_ESDA_OUTP;

	IIC0_EEnd();

	*IicData = data;
}

void IIC4_SCLH_SDAH(void)
{
	IIC4_ESCL_Hi;
	IIC4_ESDA_Hi;
	Delay();
}

void IIC4_SCLH_SDAL(void)
{
	IIC4_ESCL_Hi;
	IIC4_ESDA_Lo;
	Delay();
}

void IIC4_SCLL_SDAH(void)
{
	IIC4_ESCL_Lo;
	IIC4_ESDA_Hi;
	Delay();
}

void IIC4_SCLL_SDAL(void)
{
	IIC4_ESCL_Lo;
	IIC4_ESDA_Lo;
	Delay();
}

void IIC4_ELow(void)
{
	IIC4_SCLL_SDAL();
	IIC4_SCLH_SDAL();
	IIC4_SCLH_SDAL();
	IIC4_SCLL_SDAL();
}

void IIC4_EHigh(void)
{
	IIC4_SCLL_SDAH();
	IIC4_SCLH_SDAH();
	IIC4_SCLH_SDAH();
	IIC4_SCLL_SDAH();
}

void IIC4_EStart(void)
{
	IIC4_SCLH_SDAH();
	IIC4_SCLH_SDAL();
	Delay();
	IIC4_SCLL_SDAL();
}

void IIC4_EEnd(void)
{
	IIC4_SCLL_SDAL();
	IIC4_SCLH_SDAL();
	Delay();
	IIC4_SCLH_SDAH();
}

void IIC4_EAck_write(void)
{
	unsigned long ack;

	/* Function <- Input */
	IIC4_ESDA_INP;

	IIC4_ESCL_Lo;
	Delay();
	IIC4_ESCL_Hi;
	Delay();
	ack = GPIO4_DAT;
	IIC4_ESCL_Hi;
	Delay();
	IIC4_ESCL_Hi;
	Delay();

	/* Function <- Output (SDA) */
	IIC4_ESDA_OUTP;

	ack = (ack >> GPIO4_DAT_SHIFT) & 0x1;

	IIC4_SCLL_SDAL();
}

void IIC4_EAck_read(void)
{
	/* Function <- Output */
	IIC4_ESDA_OUTP;

	IIC4_ESCL_Lo;
	IIC4_ESCL_Lo;
	IIC4_ESDA_Hi;
	IIC4_ESCL_Hi;
	IIC4_ESCL_Hi;
	/* Function <- Input (SDA) */
	IIC4_ESDA_INP;

	IIC4_SCLL_SDAL();
}

void IIC4_ESetport(void)
{
	/* Pull Up/Down Disable	SCL, SDA */
	DIS_GPIO4_PUD;

	IIC4_ESCL_Hi;
	IIC4_ESDA_Hi;

	/* Function <- Output (SCL) */
	IIC4_ESCL_OUTP;
	/* Function <- Output (SDA) */
	IIC4_ESDA_OUTP;

	Delay();
}

void IIC4_EWrite(unsigned char ChipId,
		unsigned char IicAddr, unsigned char IicData)
{
	unsigned long i;

	IIC4_EStart();

	/* write chip id */
	for (i = 7; i > 0; i--) {
		if ((ChipId >> i) & 0x0001)
			IIC4_EHigh();
		else
			IIC4_ELow();
	}

	/* write */
	IIC4_ELow();

	/* ACK */
	IIC4_EAck_write();

	/* write reg. addr. */
	for (i = 8; i > 0; i--) {
		if ((IicAddr >> (i-1)) & 0x0001)
			IIC4_EHigh();
		else
			IIC4_ELow();
	}

	/* ACK */
	IIC4_EAck_write();

	/* write reg. data. */
	for (i = 8; i > 0; i--) {
		if ((IicData >> (i-1)) & 0x0001)
			IIC4_EHigh();
		else
			IIC4_ELow();
	}

	/* ACK */
	IIC4_EAck_write();

	IIC4_EEnd();
}

void IIC4_ERead(unsigned char ChipId,
		unsigned char IicAddr, unsigned char *IicData)
{
	unsigned long i, reg;
	unsigned char data = 0;

	IIC4_EStart();

	/* write chip id */
	for (i = 7; i > 0; i--) {
		if ((ChipId >> i) & 0x0001)
			IIC4_EHigh();
		else
			IIC4_ELow();
	}

	/* write */
	IIC4_ELow();

	/* ACK */
	IIC4_EAck_write();

	/* write reg. addr. */
	for (i = 8; i > 0; i--) {
		if ((IicAddr >> (i-1)) & 0x0001)
			IIC4_EHigh();
		else
			IIC4_ELow();
	}

	/* ACK */
	IIC4_EAck_write();

	IIC4_EStart();

	/* write chip id */
	for (i = 7; i > 0; i--) {
		if ((ChipId >> i) & 0x0001)
			IIC4_EHigh();
		else
			IIC4_ELow();
	}

	/* read */
	IIC4_EHigh();
	/* ACK */
	IIC4_EAck_write();

	/* read reg. data. */
	IIC4_ESDA_INP;

	IIC4_ESCL_Lo;
	IIC4_ESCL_Lo;
	Delay();

	for (i = 8; i > 0; i--) {
		IIC4_ESCL_Lo;
		IIC4_ESCL_Lo;
		Delay();
		IIC4_ESCL_Hi;
		IIC4_ESCL_Hi;
		Delay();
		reg = GPIO4_DAT;
		IIC4_ESCL_Hi;
		IIC4_ESCL_Hi;
		Delay();
		IIC4_ESCL_Lo;
		IIC4_ESCL_Lo;
		Delay();

		reg = (reg >> GPIO4_DAT_SHIFT) & 0x1;

		data |= reg << (i-1);
	}

	/* ACK */
	IIC4_EAck_read();
	IIC4_ESDA_OUTP;

	IIC4_EEnd();

	*IicData = data;
}

void pmic_init(void)
{
	board_id_t id = get_board_info();

	//
	// need to disable pulls downs  on MMC1 SDIO CMD and DATA lines
	// since this is conflicting with on board pull ups
	// set pins 1,3,4,5,6 to pull-down disabled
	//
	rGPC1PUD = rGPC1PUD & ~0x3FCC;

	if(id.is_dev_board || id.revision == 0x7) {
		/*
		 * Dev Board - PMIC is on IIC0
		 * (Proto FF has the same setup)
		 */
		IIC0_ESetport();

		//
		// turn on LDO8 since we need this to turn on the 32KHz clock
		// (0xc0) - turns it on
		// (0x14) - sets voltage to 1.8V (see page.125 of S2MPS11 Datasheet rev0.1)
		//
		IIC0_EWrite(0xcc, 0x44, 0xd4);

		//
		// need to wait atleast 100us before we enable the 32k clock
		// 100 ticks ~ 1us
		//
		sdelay(100 * 100);

		/* Enable 32kHz clock to AP, CP, and BT (CP is a test point) */
		IIC0_EWrite(0xcc, 0x0B, 0x07);

		/* Disable all unused LDOs to reduce self-heating */
		IIC0_EWrite(0xcc, 0x41, 0x14);
		IIC0_EWrite(0xcc, 0x42, 0x08);
		if(!id.is_dev_board) {
			// Disable LDO12 on FF units
			IIC0_EWrite(0xcc, 0x48, 0x14);
		}
		IIC0_EWrite(0xcc, 0x4B, 0x32);
		IIC0_EWrite(0xcc, 0x4c, 0x1c);
		IIC0_EWrite(0xcc, 0x4D, 0x32);
		IIC0_EWrite(0xcc, 0x4e, 0x14);
		IIC0_EWrite(0xcc, 0x4F, 0x28);
		IIC0_EWrite(0xcc, 0x50, 0x14);
		IIC0_EWrite(0xcc, 0x51, 0x14);
		IIC0_EWrite(0xcc, 0x52, 0x10);
		IIC0_EWrite(0xcc, 0x54, 0x28);
		IIC0_EWrite(0xcc, 0x55, 0x14);
		IIC0_EWrite(0xcc, 0x56, 0x2c);
		IIC0_EWrite(0xcc, 0x58, 0x28);
		IIC0_EWrite(0xcc, 0x59, 0x14);
		IIC0_EWrite(0xcc, 0x5a, 0x14);
		IIC0_EWrite(0xcc, 0x5b, 0x14);
		IIC0_EWrite(0xcc, 0x5c, 0x28);
		IIC0_EWrite(0xcc, 0x5d, 0x14);
		IIC0_EWrite(0xcc, 0x5e, 0x2c);
		IIC0_EWrite(0xcc, 0x5F, 0x10);
		IIC0_EWrite(0xcc, 0x60, 0x14);
		IIC0_EWrite(0xcc, 0x61, 0x14);
		IIC0_EWrite(0xcc, 0x62, 0x28);
		IIC0_EWrite(0xcc, 0x3b, 0x18);
	} else {
		/* FF Board later than Proto has PMIC on IIC4 (GPA2{0:1}) */
		IIC4_ESetport();

		//
		// CTRL_JIT bit in register 0x0B needs to be enabled to increase the drive strength
		// 32KHzAP_EN bit(0) is set to enable the AP clock.
		// don't enable the 32KHz clock going to the Wifi-BT module. this will be done after
		// enabling LDO8
		//
		IIC4_EWrite(0xcc, 0x0b, 0x11);


		//
		// turn on LDO8 since we need this to turn on the 32KHz clock
		// (0xc0) - turns it on
		// (0x14) - sets voltage to 1.8V (see page.125 of S2MPS11 Datasheet rev0.1)
		//
		IIC4_EWrite(0xcc, 0x44, 0xd4);
		//
		// we should wait atleast 100us before we turn on the 32KHz clock. the 300ms wait below should
		// take care of that
		//

		//
		// RDR-3817 - need to add this delay here to avoid RF module failure on some units
		// we need 600ms here to ensure that the RF module comes up
		//
		sdelay(600 * 100000); // Measured on scope - 100000 sdelay is ~1.1ms

		/* Enable 32kHz clock to AP, and BT only (CP is NC). CTRL_JIT is disabled at this point  */
		IIC4_EWrite(0xcc, 0x0B, 0x05);

		/* Disable all unused LDOs to reduce self-heating */
		IIC4_EWrite(0xcc, 0x41, 0x14);
		IIC4_EWrite(0xcc, 0x42, 0x08);
		IIC4_EWrite(0xcc, 0x48, 0x14);
		IIC4_EWrite(0xcc, 0x4B, 0x32);
		IIC4_EWrite(0xcc, 0x4c, 0x1c);
		IIC4_EWrite(0xcc, 0x4D, 0x32);
		IIC4_EWrite(0xcc, 0x4e, 0x14);
		IIC4_EWrite(0xcc, 0x4F, 0x28);
		IIC4_EWrite(0xcc, 0x50, 0x14);
		IIC4_EWrite(0xcc, 0x51, 0x14);
		IIC4_EWrite(0xcc, 0x52, 0x10);
		IIC4_EWrite(0xcc, 0x54, 0x28);
		IIC4_EWrite(0xcc, 0x55, 0x14);
		IIC4_EWrite(0xcc, 0x56, 0x2c);
		IIC4_EWrite(0xcc, 0x58, 0x28);
		IIC4_EWrite(0xcc, 0x59, 0x14);
		IIC4_EWrite(0xcc, 0x5a, 0x14);
		IIC4_EWrite(0xcc, 0x5b, 0x14);
		IIC4_EWrite(0xcc, 0x5c, 0x28);
		IIC4_EWrite(0xcc, 0x5d, 0x14);
		IIC4_EWrite(0xcc, 0x5e, 0xec);
		IIC4_EWrite(0xcc, 0x5F, 0x10);
		IIC4_EWrite(0xcc, 0x60, 0x14);
		IIC4_EWrite(0xcc, 0x61, 0x14);
		IIC4_EWrite(0xcc, 0x62, 0x28);
		IIC4_EWrite(0xcc, 0x3b, 0x18);
	}
}

void pmic_deinit(void)
{
	board_id_t id = get_board_info();

	if(id.is_dev_board || id.revision == 0x7) {
		/* CTRL1 Master Reset Disable & PWRHOLD is Low(force) */
		IIC0_EWrite(0xcc, 0x0c, 0x00);  // disable
	} else {
		IIC4_EWrite(0xcc, 0x0c, 0x00);  // disable
	}
}
