/*
 * (C) Copyright 2011 Samsung Electronics Co. Ltd
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#ifndef __PMIC_H__
#define __PMIC_H__

#define GPA2CON		*(volatile unsigned long *)(0x14010040)
#define GPA2DAT		*(volatile unsigned long *)(0x14010044)
#define GPA2PUD		*(volatile unsigned long *)(0x14010048)

#define GPB3CON		*(volatile unsigned long *)(0x140100C0)
#define GPB3DAT		*(volatile unsigned long *)(0x140100C4)
#define GPB3PUD		*(volatile unsigned long *)(0x140100C8)

#define IIC0_ESCL_Hi	GPB3DAT |= (0x1 << 1)
#define IIC0_ESCL_Lo	GPB3DAT &= ~(0x1 << 1)
#define IIC0_ESDA_Hi	GPB3DAT |= (0x1 << 0)
#define IIC0_ESDA_Lo	GPB3DAT &= ~(0x1 << 0)

#define IIC4_ESCL_Hi	GPA2DAT |= (0x1 << 1)
#define IIC4_ESCL_Lo	GPA2DAT &= ~(0x1 << 1)
#define IIC4_ESDA_Hi	GPA2DAT |= (0x1 << 0)
#define IIC4_ESDA_Lo	GPA2DAT &= ~(0x1 << 0)

#define IIC0_ESCL_INP	GPB3CON &= (~0xf << 4)
#define IIC0_ESCL_OUTP	GPB3CON = (GPB3CON & ~(0xf << 4)) | (0x1 << 4)

#define IIC4_ESCL_INP	GPA2CON &= (~0xf << 4)
#define IIC4_ESCL_OUTP	GPA2CON = (GPA2CON & ~(0xf << 4)) | (0x1 << 4)

#define IIC0_ESDA_INP	GPB3CON &= ~(0xf << 0)
#define IIC0_ESDA_OUTP	GPB3CON = (GPB3CON & ~(0xf << 0)) | (0x1 << 0)

#define IIC4_ESDA_INP	GPA2CON &= ~(0xf << 0)
#define IIC4_ESDA_OUTP	GPA2CON = (GPA2CON & ~(0xf << 0)) | (0x1 << 0)

#define GPIO0_DAT 		GPB3DAT
#define GPIO0_DAT_SHIFT  (0)
#define DIS_GPIO0_PUD	GPB3PUD &= ~(0xf << 0)

#define GPIO4_DAT 		GPA2DAT
#define GPIO4_DAT_SHIFT  (0)
#define DIS_GPIO4_PUD	GPA2PUD &= ~(0xf << 0)

/* TODO: Redo the above for GPA2[0:1] for form-factor post-proto */

#define DELAY		100
#define S2MPS11_ADDR	0xCC	// S2MPS11

#define CONFIG_PM_CALC_VOLT(x) (((x - 650) * 100) / 625)

#define ARM_EMA_CTRL_OFFSET			0x01008
#define ARM_EMA_CTRL_KFC_OFFSET			0x29008

extern void pmic_init(void);
extern void pmic_deinit(void);
extern void IIC0_ESetport(void);
extern void IIC0_EWrite(unsigned char ChipId,
		unsigned char IicAddr, unsigned char IicData);
extern void IIC0_ERead(unsigned char ChipId,
		unsigned char IicAddr, unsigned char *IicData);

extern void IIC4_ESetport(void);
extern void IIC4_EWrite(unsigned char ChipId,
		unsigned char IicAddr, unsigned char IicData);
extern void IIC4_ERead(unsigned char ChipId,
		unsigned char IicAddr, unsigned char *IicData);

/* S2MPS11 Register Address */
#define PMIC_ID		0x00
#define RTC_CTRL	0x0B
#define BUCK1_CTRL1	0x25
#define BUCK1_CTRL2	0x26
#define BUCK2_CTRL1	0x27
#define BUCK2_CTRL2	0x28
#define BUCK3_CTRL1	0x29
#define BUCK3_CTRL2	0x2A
#define BUCK4_CTRL1	0x2B
#define BUCK4_CTRL2	0x2C
#define BUCK5_CTRL1	0x2D
#define BUCK5_SW	0x2E
#define BUCK5_CTRL2	0x2F
#define BUCK5_CTRL3	0x30
#define BUCK5_CTRL4	0x31
#define BUCK5_CTRL5	0x32
#define BUCK6_CTRL1	0x33
#define BUCK6_CTRL2	0x34
#endif /*__PMIC_H__*/

